@@ -1,12 +1,12 @@
 package Net::DNS;
 
 #
-# $Id: DNS.pm 1089 2012-12-21 09:56:30Z willem $
+# $Id: DNS.pm 1124 2013-11-13 14:56:20Z willem $
 #
 use vars qw($SVNVERSION $VERSION);
 BEGIN {
-	$SVNVERSION = (qw$LastChangedRevision: 1089 $)[1];
-	$VERSION = '0.71_01';
+	$SVNVERSION = (qw$LastChangedRevision: 1124 $)[1];
+	$VERSION = '0.72_01';
 }
 
 
@@ -418,7 +418,7 @@ sub rr_add {
 
 sub rr_del {
 	my ( $head, @tail ) = split /\s+/, shift;
-	my $rr = new Net::DNS::RR( scalar @tail > 1 ? "$head @tail": "$head ANY @tail" );
+	my $rr = new Net::DNS::RR( scalar @tail ? "$head @tail": "$head ANY" );
 	$rr->ttl(0);
 	$rr->class( $rr->rdata ? 'NONE' : 'ANY' );
 	return $rr;
